import sys
from xml import etree
from xml.etree.ElementTree import Element, tostring
import os
from .metadata import MetriqlMetadata
from tableaudocumentapi import Datasource

direct_mapping_types = ['boolean', 'integer', 'boolean', 'date']


class GenerateTDS:
    metadata: MetriqlMetadata

    def __init__(self, metadata: MetriqlMetadata) -> None:
        self.metadata = metadata

    def generate(self, dataset_name, output_file: str):
        dataset = self.metadata.get_dataset(dataset_name)
        source_tds = Datasource.from_file(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'boilerplate.tds'))
        datasource_xml = source_tds._datasourceXML

        url = self.metadata.get_url()
        connection = source_tds.connections[0]
        connection.server = url.hostname
        connection.port = str(url.port)

        relation = datasource_xml.find('.//relation')
        relation.set('name', dataset.get('name'))
        relation.set('table', "[{}].[{}]".format(dataset.get('category') if dataset.get('category') else "public",
                                                 dataset.get('name')))

        dimensions = self.metadata.get_dimensions(dataset.get('name'))
        measures = self.metadata.get_measures(dataset.get('name'))
        drill_paths = self.append_columns(dataset, datasource_xml, dimensions, measures)

        # must be written just before the folders
        drill_paths_node = Element("drill-paths")
        for path in drill_paths:
            drill_paths_node.append(path)
        datasource_xml.append(drill_paths_node)

        self.append_folders(datasource_xml, dimensions, measures)

        field = Element("layout",
                        {"dim-ordering": "alphabetic", "dim-percentage": "0.5", "measure-ordering": "alphabetic",
                         "measure-percentage": "0.4", "show-structure": "false"})
        datasource_xml.append(field)

        self.indent(source_tds._datasourceXML)

        if output_file is not None:
            source_tds._datasourceTree.write(output_file, encoding="utf-8", xml_declaration=True)
        else:
            sys.stdout.buffer.write("<?xml version='1.0' encoding='utf-8'?>\n")
            sys.stdout.buffer.write(tostring(source_tds._datasourceTree._root))

    @staticmethod
    def _get_field_category(field, relation):
        category = field.get('category')
        if relation is not None:
            if category is not None:
                return "{} / {}".format(relation.get('label'), )
            else:
                return relation.get('label')
        elif category is not None:
            return category

    @staticmethod
    def append_folders(datasource_xml, dimensions, measures):
        dimension_folders = {}
        for name, (dimension, relation) in dimensions.items():
            category = GenerateTDS._get_field_category(dimension, relation)
            if category is not None:
                columns = dimension_folders.setdefault(category, [])
                columns.append(name)

        for name, columns in dimension_folders.items():
            datasource_xml.append(GenerateTDS._create_folder(name, 'dimensions', columns))

        measure_folders = {}
        for name, (measure, relation) in measures.items():
            GenerateTDS._get_field_category(measure, relation)
            if category is not None:
                names = measure_folders.setdefault(category, [])
                names.append(name)

        for name, columns in measure_folders.items():
            datasource_xml.append(GenerateTDS._create_folder(name, 'measures', columns))

    def append_columns(self, dataset, datasource_xml, dimensions, measures):
        drill_paths = []

        for name, (dimension, relation) in dimensions.items():
            post_operations = dimension.get('postOperations')
            if post_operations is None:
                self._append_column(dataset, datasource_xml, name, dimension)
            else:
                for post_operation in post_operations:
                    self._append_column(dataset, datasource_xml, name, dimension, post_operation=post_operation)
                drill_paths.append(
                    self._create_drill_path(name, map(lambda op: "{}::{}".format(name, op), post_operations)))

        for name, (measure, relation) in measures.items():
            measure_type = measure.get('type')
            measure_value = measure.get('value')
            if measure_type == 'dimension':
                pass
            elif measure_type == 'column':
                if measure_value.get('column') is not None:
                    # check if there is such dimension
                    dimension_for_measure = self.metadata.get_dimension_for_column(dataset, measure_value.get('column'))
                    if dimension_for_measure is None:
                        self._append_column(dataset, datasource_xml, name, dimension,
                                            aggregation=measure_value.get('aggregation'))
            elif measure_type == 'sql':
                aggregation = measure_value.get('aggregation')
                self._append_column(dataset, datasource_xml, name, dimension, aggregation=aggregation,
                                    formula=measure_value.get('sql'))
            else:
                raise ValueError

        return drill_paths

    @staticmethod
    def indent(elem, level=0):
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                GenerateTDS.indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    @staticmethod
    def _get_column_type(fieldType, postOperation):
        if fieldType in direct_mapping_types or postOperation is not None:
            return "ordinal"  # 1,2,3
        else:
            return "nominal"  # category

    @staticmethod
    def _get_column_datatype(field_type):
        if field_type in ['timestamp', 'date']:
            return "string"

        if field_type in ['boolean', 'integer', 'boolean', 'string']:
            return field_type

        if field_type in ['double', 'long']:
            return 'real'

        raise Exception("Unknown type {}".format(field_type))

    def _append_column(self, dataset, datasource_xml, name, field, post_operation=None, aggregation=None,
                       formula=None):
        name_reference = "{}::{}".format(name, post_operation) if post_operation is not None else name
        label = field.get('label') or name
        caption = post_operation if post_operation is not None else label

        is_measure = aggregation is not None or formula is not None
        if is_measure:
            tableau_type = "quantitative"
        else:
            tableau_type = GenerateTDS._get_column_type(field.get('fieldType'), post_operation)

        node = Element("column", {"name": "[{}]".format(name_reference),
                                  "caption": caption,
                                  "role": "dimension" if not is_measure else "measure",
                                  "default-role": "dimension" if formula is None else "measure",
                                  "type": tableau_type})

        node.attrib["datatype"] = GenerateTDS._get_column_datatype(field.get('fieldType'))

        if formula is not None:
            formula_node = Element('calculation', {"formula": formula, "class": "tableau"})
            node.append(formula_node)
        else:
            if len(field.get("description", '')) > 0:
                node.append(GenerateTDS._create_desc_node(field.get("description")))

        default_aggregation = aggregation or self.metadata.default_aggregation_for_dimension(dataset, field)
        if default_aggregation is not None:
            node.attrib['aggregation'] = self.convert_tableau_aggregation(default_aggregation)

        datasource_xml.append(node)

    @staticmethod
    def convert_tableau_aggregation(aggregation):
        if aggregation in ['count', 'sum']:
            return aggregation.capitalize()
        if aggregation in ['approximateUnique', 'countUnique']:
            return 'CountD'
        if aggregation == 'average':
            return 'Avg'
        if aggregation == 'minimum':
            return 'Min'
        if aggregation == 'maximum':
            return 'Max'
        if aggregation == 'sumDistinct':
            return 'Sum'
        if aggregation == 'averageDistinct':
            return 'Avg'
        else:
            raise Exception('Unknown aggregation {}'.format(aggregation))

    @staticmethod
    def _create_folder(name, role, columns):
        folder = Element('folder', {"name": name, "role": role})
        for column in columns:
            folder_item = Element("folder-item", {"name": '[{}]'.format(column), "type": "field"})
            folder.append(folder_item)
        return folder

    @staticmethod
    def _create_drill_path(name, columns):
        drill_path = Element('drill-path', {"name": name})
        for column in columns:
            field = Element("field")
            field.text = '[{}]'.format(column)
            drill_path.append(field)
        return drill_path

    @staticmethod
    def _create_desc_node(description):
        desc = Element('desc')
        formatted_text = Element("formatted-text")
        run = Element("run")
        run.text = description
        desc.append(formatted_text)
        formatted_text.append(run)
        return desc



from urllib.parse import urlparse, ParseResult


class MetriqlMetadata:
    _datasets: list
    _url: ParseResult

    def __init__(self, url: str, datasets):
        self._datasets = datasets
        self._url = urlparse(url)

    def get_url(self):
        return self._url

    @staticmethod
    def get_dimension_for_column(dataset, column_name):
        for dimension in dataset.get('dimensions'):
            if dimension.get('type') == 'column' and dimension.get('value').get('column') == column_name:
                return dimension

    @staticmethod
    def default_aggregation_for_dimension(dataset, dimension):
        column_reference = None
        if dimension.get('type') == 'column':
            column_reference = dimension.get('value').get('column')

        for measure in dataset.get('measures'):
            field_value = measure.get('value')
            if measure.get('type') == 'dimension':
                if field_value.get('dimension') == dimension.get('name'):
                    return field_value.get('aggregation')
            if column_reference is not None and measure.get('type') == 'column':
                if field_value.get('column') == column_reference:
                    return field_value.get('aggregation')

    def get_dimensions(self, dataset_name):
        return self._get_fields(dataset_name, 'dimensions')

    def get_measures(self, dataset_name):
        return self._get_fields(dataset_name, 'measures')

    def _get_fields(self, dataset_name, field_type):
        fields = {}
        source_dataset = self.get_dataset(dataset_name)
        self._populate_fields(fields, source_dataset.get(field_type), None)

        for relation in source_dataset.get('relations'):
            relation_dataset = self.get_dataset(relation.get('modelName'))
            self._populate_fields(fields, relation_dataset.get(field_type), relation)

        return fields

    def get_dataset(self, name):
        return next(filter(lambda d: d.get('name') == name, self._datasets))

    @staticmethod
    def _get_field_by_name(fields, name):
        return next(filter(lambda d: d.get('name') == name, fields))

    @staticmethod
    def _populate_fields(fields, dataset_fields, relation):
        field_prefix = "{}.".format(relation.get('name')) if relation is not None else ""
        for field in dataset_fields:
            fields[field_prefix + field.get('name')] = field, relation



import argparse
import json
from .metadata import MetriqlMetadata
from .generate import GenerateTDS
import sys

__version__ = "0.2"


def main(args: list = None):
    parser = argparse.ArgumentParser(
        description="Generates Tableau TDS files for metriql datasets"
    )

    parser.add_argument("command", choices=["create-tds"], help="command to execute")

    parser.add_argument("--metriql-url", help="metriql URL", )

    parser.add_argument("--dataset", help="dataset for the TDS file")

    parser.add_argument("--file", help="source of the metadata file. if not set, the source is stdin")

    parser.add_argument("--out", help="target location for TDS file. if not set, the source is stdout")

    parsed = parser.parse_args(args=args)
    if parsed.command == "create-tds":
        if parsed.file is not None:
            source = open(parsed.file).read()
        else:
            source = sys.stdin.readline()
        metriql_metadata = MetriqlMetadata(parsed.metriql_url, json.loads(source))
        GenerateTDS(metriql_metadata).generate(parsed.dataset, parsed.out)



#!/usr/bin/env python3

import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 6):
    raise ValueError("Requires Python 3.6+")

from metriql2tableau import __version__

with open("requirements.txt", "r") as f:
    requires = [x.strip() for x in f if x.strip()]

with open("requirements-test.txt", "r") as f:
    test_requires = [x.strip() for x in f if x.strip()]

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="metriql-tableau",
    version=__version__,
    description="metriql Tableau integration",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Burak Emre Kabakci",
    author_email="emre@rakam.io",
    url="https://github.com/metriql/metriql-tableau",
    license="MIT License",
    packages=find_packages(exclude=["tests"]),
    test_suite="tests",
    scripts=["metriql2tableau/bin/metriql-tableau"],
    tests_require=test_requires,
    install_requires=requires,
    extras_require={"test": test_requires},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)